@@ -87,7 +87,8 @@ sub get_init_service_name {
 sub get_init_scripts {
     my ($info) = @_;
     my @scripts;
-    if (my $initd_path = $info->index_resolved_path('etc/init.d/')) {
+    if ($info->name ne 'initscripts'
+        and my $initd_path = $info->index_resolved_path('etc/init.d/')) {
         for my $init_script ($initd_path->children) {
             # sysv generator drops the .sh suffix
             my $basename = get_init_service_name($init_script);
@@ -261,6 +262,31 @@ sub check_systemd_service_file {
           unless $seen_hardening
           or $is_oneshot
           or any { 'sleep.target' eq $_ } @wanted_by;
+
+        #<<< no perltidy
+        if (extract_service_file_values($file, 'Unit', 'DefaultDependencies', 1)) {
+        #>>>
+            my $seen_conflicts_shutdown = 0;
+            my @conflicts
+              = extract_service_file_values($file, 'Unit','Conflicts', 1);
+            foreach my $x (@conflicts) {
+                next unless $x eq 'shutdown.target';
+                $seen_conflicts_shutdown = 1;
+                last;
+            }
+            if ($seen_conflicts_shutdown) {
+                my $seen_before_shutdown = 0;
+                my @before
+                  = extract_service_file_values($file, 'Unit','Before', 1);
+                foreach my $x (@before) {
+                    next unless $x eq 'shutdown.target';
+                    $seen_before_shutdown = 1;
+                    last;
+                }
+                tag 'systemd-service-file-shutdown-problems', $file,
+                  unless $seen_before_shutdown;
+            }
+        }
     }
 
     return 1;
@@ -359,7 +385,8 @@ sub extract_service_file_values {
 sub check_timers {
     my ($info) = @_;
 
-    return if any { m,^/lib/systemd/system/\.timer$, } $info->sorted_index;
+    return
+      if any { m,^lib/systemd/system/[^\/]+\.timer$, } $info->sorted_index;
 
     for my $file ($info->sorted_index) {
         tag 'missing-systemd-timer-for-cron-script', $file
@@ -393,7 +420,7 @@ sub check_maintainer_scripts {
             next if substr($_, 0, $-[0]) =~ /#/;
 
             # systemctl should not be called in maintainer scripts at all,
-            # except for systemctl --daemon-reload calls.
+            # except for systemctl daemon-reload calls.
             if (m/^(?:.+;)?\s*systemctl\b/ && !/daemon-reload/) {
                 tag 'maintainer-script-calls-systemctl', "$file:$.";
             }
