@@ -20,6 +20,7 @@
 # MA 02110-1301, USA.
 
 package Lintian::Util;
+
 use strict;
 use warnings;
 use autodie;
@@ -28,27 +29,19 @@ use Carp qw(croak);
 use Cwd qw(abs_path);
 use Errno qw(ENOENT);
 use Exporter qw(import);
+use IO::Async::Loop;
+use IO::Async::Process;
 use Path::Tiny;
 use POSIX qw(sigprocmask SIG_BLOCK SIG_UNBLOCK SIG_SETMASK);
-use YAML::XS ();
 
-use constant {
-    DCTRL_DEBCONF_TEMPLATE => 1,
-    DCTRL_NO_COMMENTS => 2,
-    DCTRL_COMMENTS_AT_EOL => 4,
-};
+use Lintian::Deb822Parser qw(read_dpkg_control parse_dpkg_control);
 
 # Force export as soon as possible, since some of the modules we load also
 # depend on us and the sequencing can cause things not to be exported
 # otherwise.
-our (@EXPORT_OK, %EXPORT_TAGS);
+our @EXPORT_OK;
 
 BEGIN {
-    %EXPORT_TAGS
-      = (constants =>
-          [qw(DCTRL_DEBCONF_TEMPLATE DCTRL_NO_COMMENTS DCTRL_COMMENTS_AT_EOL)]
-      );
-
     eval { require PerlIO::gzip };
     if ($@) {
         *open_gz = \&__open_gz_ext;
@@ -57,10 +50,6 @@ BEGIN {
     }
 
     @EXPORT_OK = (qw(
-          visit_dpkg_paragraph
-          parse_dpkg_control
-          read_dpkg_control
-          read_dpkg_control_utf8
           get_deb_info
           get_dsc_info
           get_file_checksum
@@ -71,13 +60,15 @@ BEGIN {
           internal_error
           do_fork
           run_cmd
+          safe_qx
           strip
           lstrip
           rstrip
-          system_env
           copy_dir
+          sort_file_index
           gunzip_file
           open_gz
+          gzip
           perm2oct
           check_path
           clean_env
@@ -88,29 +79,27 @@ BEGIN {
           drain_pipe
           signal_number2name
           dequote_name
-          load_state_cache
-          save_state_cache
-          find_backlog
           pipe_tee
           untaint
           $PKGNAME_REGEX
           $PKGREPACK_REGEX
           $PKGVERSION_REGEX
-          ),
-        @{ $EXPORT_TAGS{constants} });
+    ));
 }
 
 use Digest::MD5;
 use Digest::SHA;
 use Encode ();
-use File::Temp qw(tempfile);
 use FileHandle;
-use List::Util qw(shuffle);
 use Scalar::Util qw(openhandle);
 
 use Lintian::Command qw(spawn);
 use Lintian::Relation::Version qw(versions_equal versions_comparator);
 
+use constant EMPTY => q{};
+use constant SPACE => q{ };
+use constant NEWLINE => qq{\n};
+
 =head1 NAME
 
 Lintian::Util - Lintian utility functions
@@ -123,20 +112,6 @@ Lintian::Util - Lintian utility functions
  if (defined $path) {
     # ...
  }
- 
- my (@paragraphs);
- eval { @paragraphs = read_dpkg_control_utf8('some/debian/ctrl/file'); };
- if ($@) {
-    # syntax error etc.
-    die "ctrl/file: $@";
- }
- 
- foreach my $para (@paragraphs) {
-    my $value = $para->{'some-field'};
-    if (defined $value) {
-        # ...
-    }
- }
 
 =head1 DESCRIPTION
 
@@ -145,61 +120,6 @@ but on their own did not warrant their own module.
 
 Most subs are imported only on request.
 
-=head2 Debian control parsers
-
-At first glance, this module appears to contain several debian control
-parsers.  In practise, there is only one real parser
-(L</visit_dpkg_paragraph>) - the rest are convenience functions around
-it.
-
-If you have very large files (e.g. Packages_amd64), you almost
-certainly want L</visit_dpkg_paragraph>.  Otherwise, one of the
-convenience functions are probably what you are looking for.
-
-=over 4
-
-=item Use L</get_deb_info> when
-
-You have a I<.deb> (or I<.udeb>) file and you want the control file
-from it.
-
-=item Use L</get_dsc_info> when
-
-You have a I<.dsc> (or I<.changes>) file.  Alternative, it is also
-useful if you have a control file and only care about the first
-paragraph.
-
-=item Use L</read_dpkg_control_utf8> or L</read_dpkg_control> when
-
-You have a debian control file (such I<debian/control>) and you want
-a number of paragraphs from it.
-
-=item Use L</parse_dpkg_control> when
-
-When you would have used L</read_dpkg_control_utf8>, except you have an
-open filehandle rather than a file name.
-
-=back
-
-=head1 CONSTANTS
-
-The following constants can be passed to the Debian control file
-parser functions to alter their parsing flag.
-
-=over 4
-
-=item DCTRL_DEBCONF_TEMPLATE
-
-The file should be parsed as debconf template.  These have slightly
-syntax rules for whitespace in some cases.
-
-=item DCTRL_NO_COMMENTS
-
-The file do not allow comments.  With this flag, any comment in the
-file is considered a syntax error.
-
-=back
-
 =head1 VARIABLES
 
 =over 4
@@ -220,7 +140,7 @@ not anchored and does not enforce any "boundary" characters.
 
 =cut
 
-our $PKGREPACK_REGEX = qr/(dfsg|debian|ds|repack)/o;
+our $PKGREPACK_REGEX = qr/(dfsg|debian|ds|repack)[^-]*\-[^-]+$/o;
 
 =item $PKGVERSION_REGEX
 
@@ -242,432 +162,6 @@ our $PKGVERSION_REGEX = qr/
 
 =over 4
 
-=item parse_dpkg_control(HANDLE[, FLAGS[, LINES]])
-
-Reads a debian control file from HANDLE and returns a list of
-paragraphs in it.  A paragraph is represented via a hashref, which
-maps (lower cased) field names to their values.
-
-FLAGS (if given) is a bitmask of the I<DCTRL_*> constants.  Please
-refer to L</CONSTANTS> for the list of constants and their meaning.
-The default value for FLAGS is 0.
-
-If LINES is given, it should be a reference to an empty list.  On
-return, LINES will be populated with a hashref for each paragraph (in
-the same order as the returned list).  Each hashref will also have a
-special key "I<START-OF-PARAGRAPH>" that gives the line number of the
-first field in that paragraph.  These hashrefs will map the field name
-of the given paragraph to the line number where the field name
-appeared.
-
-This is a convenience sub around L</visit_dpkg_paragraph> and can
-therefore produce the same errors as it.  Please see
-L</visit_dpkg_paragraph> for the finer semantics of how the
-control file is parsed.
-
-NB: parse_dpkg_control does I<not> close the handle for the caller.
-
-=cut
-
-sub parse_dpkg_control {
-    my ($handle, $flags, $lines) = @_;
-    my @result;
-    my $c = sub {
-        my ($para, $line) = @_;
-        push @result, $para;
-        push @$lines, $line if defined $lines;
-    };
-    visit_dpkg_paragraph($c, $handle, $flags);
-    return @result;
-}
-
-=item visit_dpkg_paragraph (CODE, HANDLE[, FLAGS])
-
-Reads a debian control file from HANDLE and passes each paragraph to
-CODE.  A paragraph is represented via a hashref, which maps (lower
-cased) field names to their values.
-
-FLAGS (if given) is a bitmask of the I<DCTRL_*> constants.  Please
-refer to L</CONSTANTS> for the list of constants and their meaning.
-The default value for FLAGS is 0.
-
-If the file is empty (i.e. it contains no paragraphs), the method will
-contain an I<empty> list.  The deb822 contents may be inside a
-I<signed> PGP message with a signature.
-
-visit_dpkg_paragraph will require the PGP headers to be correct (if
-present) and require that the entire file is covered by the signature.
-However, it will I<not> validate the signature (in fact, the contents
-of the PGP SIGNATURE part can be empty).  The signature should be
-validated separately.
-
-visit_dpkg_paragraph will pass paragraphs to CODE as they are
-completed.  If CODE can process the paragraphs as they are seen, very
-large control files can be processed without keeping all the
-paragraphs in memory.
-
-As a consequence of how the file is parsed, CODE may be passed a
-number of (valid) paragraphs before parsing is stopped due to a syntax
-error.
-
-NB: visit_dpkg_paragraph does I<not> close the handle for the caller.
-
-CODE is expected to be a callable reference (e.g. a sub) and will be
-invoked as the following:
-
-=over 4
-
-=item CODE->(PARA, LINE_NUMBERS)
-
-The first argument, PARA, is a hashref to the most recent paragraph
-parsed.  The second argument, LINE_NUMBERS, is a hashref mapping each
-of the field names to the line number where the field name appeared.
-LINE_NUMBERS will also have a special key "I<START-OF-PARAGRAPH>" that
-gives the line number of the first field in that paragraph.
-
-The return value of CODE is ignored.
-
-If the CODE invokes die (or similar) the error is propagated to the
-caller.
-
-=back
-
-
-I<On syntax errors>, visit_dpkg_paragraph will call die with the
-following string:
-
-  "syntax error at line %d: %s\n"
-
-Where %d is the line number of the issue and %s is one of:
-
-=over
-
-=item Duplicate field %s
-
-The field appeared twice in the paragraph.
-
-=item Continuation line outside a paragraph (maybe line %d should be " .")
-
-A continuation line appears outside a paragraph - usually caused by an
-unintended empty line before it.
-
-=item Whitespace line not allowed (possibly missing a ".")
-
-An empty continuation line was found.  This usually means that a
-period is missing to denote an "empty line" in (e.g.) the long
-description of a package.
-
-=item Cannot parse line "%s"
-
-Generic error containing the text of the line that confused the
-parser.  Note that all non-printables in %s will be replaced by
-underscores.
-
-=item Comments are not allowed
-
-A comment line appeared and FLAGS contained DCTRL_NO_COMMENTS.
-
-=item PGP signature seen before start of signed message
-
-A "BEGIN PGP SIGNATURE" header is seen and a "BEGIN PGP MESSAGE" has
-not been seen yet.
-
-=item Two PGP signatures (first one at line %d)
-
-Two "BEGIN PGP SIGNATURE" headers are seen in the same file.
-
-=item Unexpected %s header
-
-A valid PGP header appears (e.g. "BEGIN PUBLIC KEY BLOCK").
-
-=item Malformed PGP header
-
-An invalid or malformed PGP header appears.
-
-=item Expected at most one signed message (previous at line %d)
-
-Two "BEGIN PGP MESSAGE" headers appears in the same message.
-
-=item End of file but expected an "END PGP SIGNATURE" header
-
-The file ended after a "BEGIN PGP SIGNATURE" header without being
-followed by an "END PGP SIGNATURE".
-
-=item PGP MESSAGE header must be first content if present
-
-The file had content before PGP MESSAGE.
-
-=item Data after the PGP SIGNATURE
-
-The file had data after the PGP SIGNATURE block ended.
-
-=item End of file before "BEGIN PGP SIGNATURE"
-
-The file had a "BEGIN PGP MESSAGE" header, but no signature was
-present.
-
-=back
-
-=cut
-
-sub visit_dpkg_paragraph {
-    my ($code, $CONTROL, $flags) = @_;
-    $flags//=0;
-    my $lines = {};
-    my $section = {};
-    my $open_section = 0;
-    my $last_tag;
-    my $debconf = $flags & DCTRL_DEBCONF_TEMPLATE;
-    my $signed = 0;
-    my $signature = 0;
-
-    local $_;
-    while (<$CONTROL>) {
-        chomp;
-
-        if (substr($_, 0, 1) eq '#') {
-            next unless $flags & DCTRL_NO_COMMENTS;
-            die "syntax error at line $.: Comments are not allowed.\n";
-        }
-
-        # empty line?
-        if ($_ eq '' || (!$debconf && m/^\s*$/)) {
-            if ($open_section) { # end of current section
-                 # pass the current section to the handler
-                $code->($section, $lines);
-                $section = {};
-                $lines = {};
-                $open_section = 0;
-            }
-        }
-        # pgp sig? Be strict here (due to #696230)
-        # According to http://tools.ietf.org/html/rfc4880#section-6.2
-        # The header MUST start at the beginning of the line and MUST NOT have
-        # any other text (except whitespace) after the header.
-        elsif (m/^-----BEGIN PGP SIGNATURE-----[ \r\t]*$/)
-        { # skip until end of signature
-            my $saw_end = 0;
-            if (not $signed or $signature) {
-                die join(q{ },
-                    "syntax error at line $.:",
-                    "PGP signature seen before start of signed message\n")
-                  if not $signed;
-                die join(q{ },
-                    "syntax error at line $.:",
-                    "Two PGP signatures (first one at line $signature)\n");
-            }
-            $signature = $.;
-            while (<$CONTROL>) {
-                if (m/^-----END PGP SIGNATURE-----[ \r\t]*$/o) {
-                    $saw_end = 1;
-                    last;
-                }
-            }
-            # The "at line X" may seem a little weird, but it keeps the
-            # message format identical.
-            die join(q{ },
-                "syntax error at line $.:",
-                qq{End of file but expected an "END PGP SIGNATURE" header\n})
-              unless $saw_end;
-        }
-        # other pgp control?
-        elsif (m/^-----(?:BEGIN|END) PGP/) {
-            # At this point it could be a malformed PGP header or one
-            # of the following valid headers (RFC4880):
-            #  * BEGIN PGP MESSAGE
-            #    - Possibly a signed Debian CTRL, so okay (for now)
-            #  * BEGIN PGP {PUBLIC,PRIVATE} KEY BLOCK
-            #    - Valid header, but not a Debian CTRL file.
-            #  * BEGIN PGP MESSAGE, PART X{,/Y}
-            #    - Valid, but we don't support partial messages, so
-            #      bail on those.
-
-            unless (m/^-----BEGIN PGP SIGNED MESSAGE-----[ \r\t]*$/) {
-                # Not a (full) PGP MESSAGE; reject.
-
-                my $key = qr/(?:BEGIN|END) PGP (?:PUBLIC|PRIVATE) KEY BLOCK/;
-                my $msgpart = qr{BEGIN PGP MESSAGE, PART \d+(?:/\d+)?};
-                my $msg
-                  = qr/(?:BEGIN|END) PGP (?:(?:COMPRESSED|ENCRYPTED) )?MESSAGE/;
-
-                if (m/^-----($key|$msgpart|$msg)-----[ \r\t]*$/o) {
-                    die "syntax error at line $.: Unexpected $1 header\n";
-                } else {
-                    die "syntax error at line $.: Malformed PGP header\n";
-                }
-            } else {
-                if ($signed) {
-                    die join(q{ },
-                        "syntax error at line $.:",
-                        'Expected at most one signed message',
-                        "(previous at line $signed)\n");
-                }
-                if ($last_tag) {
-                    # NB: If you remove this, keep in mind that it may
-                    # allow two paragraphs to merge.  Consider:
-                    #
-                    # Field-P1: some-value
-                    # -----BEGIN PGP SIGNATURE-----
-                    #
-                    # Field-P2: another value
-                    #
-                    # At the time of writing: If $open_section is
-                    # true, it will remain so until the empty line
-                    # after the PGP header.
-                    die join(q{ },
-                        "syntax error at line $.:",
-                        'PGP MESSAGE header must be first',
-                        "content if present\n");
-                }
-                $signed = $.;
-            }
-
-            # skip until the next blank line
-            while (<$CONTROL>) {
-                last if /^\s*$/o;
-            }
-        }
-       # did we see a signature already?  We allow all whitespace/comment lines
-       # outside the signature.
-        elsif ($signature) {
-            # Accept empty lines after the signature.
-            next if m/^\s*$/;
-
-            # NB: If you remove this, keep in mind that it may allow
-            # two paragraphs to merge.  Consider:
-            #
-            # Field-P1: some-value
-            # -----BEGIN PGP SIGNATURE-----
-            # [...]
-            # -----END PGP SIGNATURE-----
-            # Field-P2: another value
-            #
-            # At the time of writing: If $open_section is true, it
-            # will remain so until the empty line after the PGP
-            # header.
-            die "syntax error at line $.: Data after the PGP SIGNATURE\n";
-        }
-        # new empty field?
-        elsif (m/^([^: \t]+):\s*$/o) {
-            $lines->{'START-OF-PARAGRAPH'} = $. if not $open_section;
-            $open_section = 1;
-
-            my ($tag) = (lc $1);
-            $section->{$tag} = '';
-            $lines->{$tag} = $.;
-
-            $last_tag = $tag;
-        }
-        # new field?
-        elsif (m/^([^: \t]+):\s*(.*)$/o) {
-            $lines->{'START-OF-PARAGRAPH'} = $. if not $open_section;
-            $open_section = 1;
-
-            # Policy: Horizontal whitespace (spaces and tabs) may occur
-            # immediately before or after the value and is ignored there.
-            my ($tag,$value) = (lc $1,$2);
-            rstrip($value);
-            if (exists $section->{$tag}) {
-                # Policy: A paragraph must not contain more than one instance
-                # of a particular field name.
-                die "syntax error at line $.: Duplicate field $tag.\n";
-            }
-            $value =~ s/#.*$// if $flags & DCTRL_COMMENTS_AT_EOL;
-            $section->{$tag} = $value;
-            $lines->{$tag} = $.;
-
-            $last_tag = $tag;
-        }
-        # continued field?
-        elsif (m/^([ \t].*\S.*)$/o) {
-            $open_section
-              or die join(q{ },
-                "syntax error at line $.:",
-                'Continuation line outside a paragraph (maybe line',
-                ($. - 1), qq{should be " .").\n});
-
-            # Policy: Many fields' values may span several lines; in this case
-            # each continuation line must start with a space or a tab.  Any
-            # trailing spaces or tabs at the end of individual lines of a
-            # field value are ignored.
-            my $value = rstrip($1);
-            $value =~ s/#.*$// if $flags & DCTRL_COMMENTS_AT_EOL;
-            $section->{$last_tag} .= "\n" . $value;
-        }
-        # None of the above => syntax error
-        else {
-            my $message = "syntax error at line $.";
-            if (m/^\s+$/) {
-                $message
-                  .= ": Whitespace line not allowed (possibly missing a \".\").\n";
-            } else {
-                # Replace non-printables and non-space characters with
-                # "_" - just in case.
-                s/[^[:graph:][:space:]]/_/go;
-                $message .= ": Cannot parse line \"$_\"\n";
-            }
-            die $message;
-        }
-    }
-    # pass the last section (if not already done).
-    $code->($section, $lines) if $open_section;
-
-    # Given the API, we cannot use this check to prevent any
-    # paragraphs from being emitted to the code argument, so we might
-    # as well just do this last.
-    if ($signed and not $signature) {
-        # The "at line X" may seem a little weird, but it keeps the
-        # message format identical.
-        die join(q{ },
-            "syntax error at line $.:",
-            qq{End of file before "BEGIN PGP SIGNATURE"\n"});
-    }
-}
-
-=item read_dpkg_control_utf8(FILE[, FLAGS[, LINES]])
-
-=item read_dpkg_control(FILE[, FLAGS[, LINES]])
-
-This is a convenience function to ease using L</parse_dpkg_control>
-with paths to files (rather than open handles).  The first argument
-must be the path to a FILE, which should be read as a debian control
-file.  If the file is empty, an empty list is returned.
-
-Otherwise, this behaves like:
-
- use autodie;
- 
- open(my $fd, '<:encoding(UTF-8)', FILE); # or '<'
- my @p = parse_dpkg_control($fd, FLAGS, LINES);
- close($fd);
- return @p;
-
-This goes without saying that may fail with any of the messages that
-L</parse_dpkg_control(HANDLE[, FLAGS[, LINES]])> do.  It can also emit
-autodie exceptions if open or close fails.
-
-=cut
-
-sub read_dpkg_control {
-    my ($file, $flags, $lines) = @_;
-
-    open(my $CONTROL, '<', $file);
-    my @data = parse_dpkg_control($CONTROL, $flags, $lines);
-    close($CONTROL);
-
-    return @data;
-}
-
-sub read_dpkg_control_utf8 {
-    my ($file, $flags, $lines) = @_;
-
-    open(my $CONTROL, '<:encoding(UTF-8)', $file);
-    my @data = parse_dpkg_control($CONTROL, $flags, $lines);
-    close($CONTROL);
-
-    return @data;
-}
-
 =item get_deb_info(DEBFILE)
 
 Extracts the control file from DEBFILE and returns it as a hashref.
@@ -692,6 +186,8 @@ L</parse_dpkg_control> do.  It can also emit:
 
 {
 
+    my $loaded = 0;
+
     sub get_deb_info {
         my ($file) = @_;
 
@@ -701,7 +197,12 @@ L</parse_dpkg_control> do.  It can also emit:
             pipe_out => FileHandle->new
         };
 
-        spawn(
+        if (not $loaded) {
+            $loaded++;
+            require Lintian::Command;
+        }
+
+        Lintian::Command::spawn(
             $opts, ['dpkg-deb', '--ctrl-tarfile', $file],
             '|', ['tar', '--wildcards', '-xO', '-f', '-', '*control']);
         my @data = parse_dpkg_control($opts->{pipe_out});
@@ -893,26 +394,6 @@ sub do_fork() {
     return $pid;
 }
 
-=item system_env (CMD)
-
-Behaves like system (CMD) except that the environment of CMD is
-cleaned (as defined by L</clean_env>(1)).
-
-=cut
-
-sub system_env {
-    my $pid = do_fork;
-    if (not defined $pid) {
-        return -1;
-    } elsif ($pid == 0) {
-        clean_env(1);
-        exec @_ or die("exec of $_[0] failed: $!\n");
-    } else {
-        waitpid $pid, 0;
-        return $?;
-    }
-}
-
 =item clean_env ([CLOC])
 
 Destructively cleans %ENV - removes all variables %ENV except a
@@ -1090,6 +571,44 @@ sub run_cmd {
     return 1;
 }
 
+=item C<safe_qx(@cmd)>
+
+Emulates the C<qx()> operator by returning the captured output
+just like Capture::Tiny;
+
+Examples:
+
+  # Capture the output of a simple command
+  my $output = safe_qx('grep', 'some-pattern', 'path/to/file');
+
+=cut
+
+sub safe_qx {
+    my @command = @_;
+
+    my $loop = IO::Async::Loop->new;
+    my $future = $loop->new_future;
+
+    $loop->run_child(
+        command => [@command],
+        on_finish => sub {
+            my ($pid, $exitcode, $stdout, $stderr) = @_;
+            my $status = ($exitcode >> 8);
+
+            if ($status) {
+                my $message = "Command @command exited with status $status";
+                $message .= ": $stderr" if length $stderr;
+                $future->fail($message);
+                return;
+            }
+
+            $future->done($stdout);
+        });
+
+    # will raise an exception in case of failure
+    return $future->get;
+}
+
 =item copy_dir (ARGS)
 
 Convenient way of calling I<cp -a ARGS>.
@@ -1103,6 +622,28 @@ sub copy_dir {
     return run_cmd('cp', '-a', '--reflink=auto', '--', @_);
 }
 
+=item sort_file_index (STRING)
+
+Sorts the file index data given in STRING and returns the sorted
+result, also in a string.
+
+=cut
+
+sub sort_file_index {
+    my ($input) = @_;
+
+    my @unsorted = split(NEWLINE, $input);
+
+    # sorts according to LC_ALL=C
+    my @sorted
+      = sort { (split(SPACE, $a))[5] cmp(split(SPACE, $b))[5] } @unsorted;
+
+    my $output = EMPTY;
+    $output .= $_ . NEWLINE for @sorted;
+
+    return $output;
+}
+
 =item gunzip_file (IN, OUT)
 
 Decompresses contents of the file IN and stores the contents in the
@@ -1112,8 +653,32 @@ will cause a trappable error.
 =cut
 
 sub gunzip_file {
-    my ($in, $out) = @_;
-    run_cmd({out => $out}, 'gzip', '-dc', $in);
+    my ($inpath, $outpath) = @_;
+
+    my $loop = IO::Async::Loop->new;
+    my $future = $loop->new_future;
+
+    my @command = ('gzip', '--decompress', '--stdout', $inpath);
+    $loop->run_child(
+        command => [@command],
+        on_finish => sub {
+            my ($pid, $exitcode, $stdout, $stderr) = @_;
+            my $status = ($exitcode >> 8);
+
+            if ($status) {
+                my $message = "Command @command exited with status $status";
+                $message .= ": $stderr" if length $stderr;
+                $future->fail($message);
+                return;
+            }
+
+            path($outpath)->spew($stdout);
+            $future->done;
+        });
+
+    # will raise an exception in case of failure
+    $future->get;
+
     return;
 }
 
@@ -1142,6 +707,55 @@ sub __open_gz_ext {
     return $fd;
 }
 
+=item gzip (DATA, PATH)
+
+Compresses DATA using gzip and stores result in file located at PATH.
+
+=cut
+
+sub gzip {
+    my ($data, $path) = @_;
+
+    unlink($path)
+      if -e $path;
+
+    $data //= EMPTY;
+
+    my $loop = IO::Async::Loop->new;
+    my $future = $loop->new_future;
+    my $compressed;
+    my $stderr;
+
+    my @command = ('gzip', '--best', '--no-name', '--stdout');
+    my $process = IO::Async::Process->new(
+        command => [@command],
+        stdin => { from => $data },
+        stdout => { into => \$compressed },
+        stderr => { into => \$stderr },
+        on_finish => sub {
+            my ($self, $exitcode) = @_;
+            my $status = ($exitcode >> 8);
+
+            if ($status) {
+                my $message = "Command @command exited with status $status";
+                $message .= ": $stderr" if length $stderr;
+                $future->fail($message);
+                return;
+            }
+
+            $future->done('Done with command @command');
+            return;
+        });
+
+    $loop->add($process);
+
+    $future->get;
+
+    path($path)->spew($compressed // EMPTY);
+
+    return;
+}
+
 =item internal_error (MSG[, ...])
 
 Use to signal an internal error. The argument(s) will used to print a
@@ -1282,14 +896,9 @@ sub lstrip (_) { ## no critic (Subroutines::RequireFinalReturn)
     # void context, so no return needed here.
 }
 
-# prototype for default to $_
-sub rstrip (_) {  ## no critic (Subroutines::RequireFinalReturn)
-    if (defined wantarray) {
-        # unpack 'A*' is faster than s/\s++$//
-        return unpack('A*', $_[0]);
-    }
-    $_[0] = unpack('A*', $_[0]);
-    # void context, so no return needed here.
+{
+    no warnings qw(once);
+    *rstrip = \&Lintian::Deb822Parser::rstrip;
 }
 
 =item check_path (CMD)
@@ -1586,127 +1195,6 @@ sub pipe_tee {
     return 1;
 }
 
-=item load_state_cache(STATE_DIR)
-
-[Reporting tools only] Load the state cache from STATE_DIR.
-
-=cut
-
-sub load_state_cache {
-    my ($state_dir) = @_;
-    my $state_file = "$state_dir/state-cache";
-    my $state = {};
-
-    return $state
-      unless -f $state_file;
-
-    my $yaml = path($state_file)->slurp;
-
-    eval {$state = YAML::XS::Load($yaml);};
-    # Not sure what Load does in case of issues; perldoc YAML says
-    # very little about it.  Based on YAML::Error, I guess it will
-    # write stuff to STDERR and use die/croak, but it remains a
-    # guess.
-    if (my $err = $@) {
-        die("$state_file was invalid; please fix or remove it.\n$err");
-    }
-    $state //= {};
-
-    if (ref($state) ne 'HASH') {
-        die("$state_file was invalid; please fix or remove it.");
-    }
-    return $state;
-}
-
-=item save_state_cache(STATE_DIR, STATE)
-
-[Reporting tools only] Save the STATE cache to STATE_DIR.
-
-=cut
-
-sub save_state_cache {
-    my ($state_dir, $state) = @_;
-    my $state_file = "$state_dir/state-cache";
-    my ($tmp_fd, $tmp_path);
-
-    ($tmp_fd, $tmp_path) = tempfile('state-cache-XXXXXX', DIR => $state_dir);
-    ## TODO: Should tmp_fd be binmode'd as we use YAML::XS?
-
-    # atomic replacement of the state file; not a substitute for
-    # proper locking, but it will at least ensure that the file
-    # is in a consistent state.
-    eval {
-        print {$tmp_fd} YAML::XS::Dump($state);
-
-        close($tmp_fd) or die("close $tmp_path: $!");
-
-        # There is no secret in this.  Set it to 0644, so it does not
-        # require sudo access on lintian.d.o to read the file.
-        chmod(0644, $tmp_path);
-
-        rename($tmp_path, $state_file)
-          or die("rename $tmp_path -> $state_file: $!");
-    };
-    if (my $err = $@) {
-        if (-e $tmp_path) {
-            # Ignore error as we have a more important one
-            no autodie qw(unlink);
-            unlink($tmp_path);
-        }
-        die($err);
-    }
-    return 1;
-}
-
-=item find_backlog(LINTIAN_VERSION, STATE)
-
-[Reporting tools only] Given the current lintian version and the
-harness state, return a list of group ids that are part of the
-backlog.  The list is sorted based on what version of Lintian
-processed the package.
-
-Note the result is by design not deterministic to reduce the
-risk of all large packages being in the same run (e.g. like
-gcc-5 + gcc-5-cross + gcc-6 + gcc-6-cross).
-
-=cut
-
-sub find_backlog {
-    my ($lintian_version, $state) = @_;
-    my (@backlog, %by_version, @low_priority);
-    for my $group_id (keys(%{$state->{'groups'}})) {
-        my $last_version = '0';
-        my $group_data = $state->{'groups'}{$group_id};
-        my $is_out_of_date;
-        # Does this group repeatedly fail with the current version
-        # of lintian?
-        if (    exists($group_data->{'processing-errors'})
-            and $group_data->{'processing-errors'} > 2
-            and exists($group_data->{'last-error-by'})
-            and $group_data->{'last-error-by'} ne $lintian_version) {
-            # To avoid possible "starvation", we will give lower priority
-            # to packages that repeatedly fail.  They will be retried as
-            # the backlog is cleared.
-            push(@low_priority, $group_id);
-            next;
-        }
-        if (exists($group_data->{'out-of-date'})) {
-            $is_out_of_date = $group_data->{'out-of-date'};
-        }
-        if (exists($group_data->{'last-processed-by'})) {
-            $last_version = $group_data->{'last-processed-by'};
-        }
-        $is_out_of_date = 1
-          if not versions_equal($last_version, $lintian_version);
-        push(@{$by_version{$last_version}}, $group_id) if $is_out_of_date;
-    }
-    for my $v (sort(versions_comparator keys(%by_version))) {
-        push(@backlog, shuffle(@{$by_version{$v}}));
-    }
-    push(@backlog, shuffle(@low_priority)) if @low_priority;
-    return @backlog;
-}
-
 =item untaint(VALUE)
 
 Untaint VALUE
